from .common import Path, _handle_path
from .firefox import Firefox

# seems to match firefox schema well enough for all of our usage
class Waterfox(Firefox):
    @classmethod
    def data_directory(cls) -> Path:
        return _handle_path(
            {
                "linux": "~/.waterfox/",
                "darwin": "~/Library/Application Support/Waterfox/Profiles/",
            },
            browser_name=cls.__name__,
        )



from .common import (
    Path,
    Schema,
    Iterator,
    Visit,
    Metadata,
    unquote,
    _from_datetime_microseconds,
    _execute_query,
    _handle_path,
    PathIshOrConn,
)
from .firefox import Firefox

# uses a similar directory structure to firefox
class Palemoon(Firefox):
    # only major difference seems to be it doesn't have
    # the 'moz_meta' table? everyhing else is similar
    # to firefox schema
    detector = "moz_historyvisits"

    # seems to store less info that firefox schema
    # no description or preview_image
    schema = Schema(
        cols=[
            "P.url",
            "V.visit_date",
            "P.title",
        ],
        where="FROM moz_historyvisits as V, moz_places as P WHERE V.place_id = P.id",
    )

    @classmethod
    def extract_visits(cls, path: PathIshOrConn) -> Iterator[Visit]:
        for row in _execute_query(path, cls.schema.query):
            yield Visit(
                url=unquote(row["url"]),
                dt=_from_datetime_microseconds(row["visit_date"]),
                metadata=Metadata.make(title=row["title"]),
            )

    # seems the non-linux community is pretty small?
    # https://forum.palemoon.org/viewforum.php?f=41
    # no easy to way to install except to build from source
    # if someone is actually using this on mac, feel free to make an issue
    @classmethod
    def data_directory(cls) -> Path:
        return _handle_path(
            {
                "linux": "~/.moonchild productions/pale moon/",
            },
            browser_name=cls.__name__,
        )



from .common import (
    Path,
    _handle_path,
)


from .chrome import Chrome


class Chromium(Chrome):
    @classmethod
    def data_directory(cls) -> Path:
        return _handle_path(
            {
                "linux": "~/.config/chromium/",
                "darwin": "~/Library/Application Support/Chromium/",
            },
            browser_name=cls.__name__,
        )



import sys
import logging
import json as jsn
from typing import List, Optional, Sequence, Callable, Iterator

import click
import IPython  # type: ignore[import]

from .model import Visit
from .save import backup_history, _path_backup
from .browsers.all import DEFAULT_BROWSERS
from .merge import read_and_merge
from .demo import demo_visit
from .log import setup

# target for python3 -m browserexport and console_script using click
@click.group()
@click.option("--debug", is_flag=True, default=False, help="Increase log verbosity")
def cli(debug: bool) -> None:
    """
    Backup and merge your browser history!
    """
    if debug:
        setup(logging.DEBUG)


browser_names = [b.__name__.lower() for b in DEFAULT_BROWSERS]


@cli.command()
@click.option(
    "-b",
    "--browser",
    type=click.Choice(browser_names, case_sensitive=False),
    required=True,
    help="Provide browser to backup, or specify directly with --path",
)
@click.option(
    "-p",
    "--profile",
    type=str,
    default="*",
    help="Use to pick the correct profile to back up. If unspecified, will assume a single profile",
)
@click.option(
    "--path",
    type=click.Path(exists=True),
    default=None,
    help="Specify a direct path to a database to back up",
)
@click.option(
    "-t",
    "--to",
    type=click.Path(exists=True),
    default=".",
    help="Directory to store backup to",
)
def save(browser: str, profile: str, to: str, path: Optional[str]) -> None:
    """
    Backs up a current browser database file
    """
    if path is not None:
        # TODO: add profile to do a basic glob/make --path easier to use?
        # could be confusing. for now, forcing the user to specify full path
        # hopefully this isn't needed a lot/can be replaced by
        # additional Browser+platform specific default paths
        _path_backup(path, to)
        return
    backup_history(browser, to, profile=profile)


def _handle_merge(dbs: List[str], *, json: bool, stream: bool) -> None:
    ivis: Iterator[Visit] = read_and_merge(dbs)
    if json or stream:
        if stream:
            for v in ivis:
                sys.stdout.write(jsn.dumps(v.serialize()))
            sys.stdout.flush()
        else:
            click.echo(jsn.dumps([v.serialize() for v in ivis]))
    else:
        vis: List[Visit] = list(ivis)
        demo_visit(vis)
        header = f"Use {click.style('vis', fg='green')} to access visit data"
        IPython.embed(header=header)


SHARED = [
    click.option(
        "-j",
        "--json",
        is_flag=True,
        default=False,
        required=False,
        help="Print result to STDOUT as JSON",
    ),
    click.option(
        "-s",
        "--stream",
        is_flag=True,
        default=False,
        required=False,
        help="Stream JSON objects instead of printing a JSON list",
    ),
]


# decorator to apply shared arguments to inspect/merge
def shared_options(func: Callable[..., None]) -> Callable[..., None]:
    for decorator in SHARED:
        func = decorator(func)
    return func


@cli.command()
@click.argument("SQLITE_DB", type=click.Path(exists=True), required=True)
@shared_options
def inspect(sqlite_db: str, json: bool, stream: bool) -> None:
    """
    Extracts visits from a single sqlite database

    \b
    Provide a history database as the first argument
    Drops you into a REPL to access the data
    """
    _handle_merge([sqlite_db], json=json, stream=stream)


@cli.command()
@click.argument("SQLITE_DB", type=click.Path(exists=True), nargs=-1, required=True)
@shared_options
def merge(sqlite_db: Sequence[str], json: bool, stream: bool) -> None:
    """
    Extracts visits from multiple sqlite databases

    \b
    Provide multiple sqlite databases as positional arguments, e.g.:
    browserexport merge ~/data/firefox/*.sqlite

    Drops you into a REPL to access the data
    """
    _handle_merge(list(sqlite_db), json=json, stream=stream)


if __name__ == "__main__":
    cli(prog_name="browserexport")



from .common import (
    Iterator,
    Visit,
    Metadata,
    PathIshOrConn,
    Browser,
    unquote,
    Path,
    Schema,
    _execute_query,
    _from_datetime_microseconds,
    _func_if_some,
    _handle_glob,
    _handle_path,
)


class Firefox(Browser):
    detector = "moz_meta"
    schema = Schema(
        cols=[
            "P.url",
            "V.visit_date",
            "P.title",
            "P.description",
            "P.preview_image_url",
        ],
        where="FROM moz_historyvisits as V, moz_places as P WHERE V.place_id = P.id",
    )

    @classmethod
    def extract_visits(cls, path: PathIshOrConn) -> Iterator[Visit]:
        for row in _execute_query(path, cls.schema.query):
            yield Visit(
                url=unquote(row["url"]),
                dt=_from_datetime_microseconds(row["visit_date"]),
                metadata=Metadata.make(
                    title=row["title"],
                    description=row["description"],
                    preview_image=_func_if_some(row["preview_image_url"], unquote),
                ),
            )

    @classmethod
    def data_directory(cls) -> Path:
        return _handle_path(
            {
                "linux": "~/.mozilla/firefox/",
                "darwin": "~/Library/Application Support/Firefox/Profiles/",
            },
            browser_name=cls.__name__,
        )

    @classmethod
    def locate_database(cls, profile: str = "*") -> Path:
        dd: Path = cls.data_directory()
        return _handle_glob(dd, profile + "/places.sqlite")



# ununsed imports here, to bring them into scope for other files
import sys
import sqlite3
import warnings

from pathlib import Path
from urllib.parse import unquote
from datetime import datetime, timezone
from typing import List, Iterator, Optional, NamedTuple, Dict
from dataclasses import dataclass

from ..log import logger
from ..model import Visit, Metadata
from ..common import PathIsh, PathIshOrConn, _func_if_some, expand_path
from ..sqlite import _execute_query


@dataclass
class Schema:
    cols: List[str]
    where: str

    @property
    def query(self) -> str:
        return f"SELECT {', '.join(self.cols)} {self.where}"


@dataclass
class Browser:
    schema: Schema  # used to create the query to extract visit from database

    detector: str  # semi-unique name of table present on database to detect this type

    @classmethod
    def detect(cls, path: PathIshOrConn) -> bool:
        """
        Run the detector against the given path/connection to detect if the current Browser matches the schema
        """
        detector_query = f"SELECT * FROM {cls.detector}"
        logger.debug(f"{cls.__name__}: Running detector query '{detector_query}'")
        try:
            list(_execute_query(path, detector_query))
            return True
        except sqlite3.OperationalError as sql_err:
            logger.debug(str(sql_err))
            return False

    @classmethod
    def extract_visits(cls, path: PathIshOrConn) -> Iterator[Visit]:
        """
        Given a path or a sqlite3 connection, extract visits
        """
        raise NotImplementedError

    @classmethod
    def data_directory(cls) -> Path:
        """
        The local data directory for this browser
        """
        raise NotImplementedError

    @classmethod
    def locate_database(cls, profile: str) -> Path:
        """
        Locate this database on the users' computer so it can be backed up
        """
        raise NotImplementedError


def _from_datetime_microseconds(ts: int) -> datetime:
    return datetime.fromtimestamp(ts / 1_000_000, tz=timezone.utc)


errmsg = """Expected to match single database, got {}.
You can use the --profile argument to select one of the profiles/match a particular file"""


def _handle_glob(base: Path, stem: str, recursive: bool = False) -> Path:
    dbs: List[Path]
    if not recursive:
        # basic glob, in the same directory
        dbs = list(base.glob(stem))
    else:
        # recursively glob
        dbs = list(base.rglob(stem))
    recur_desc = "recursive" if recursive else "non recursive"
    logger.debug(f"Glob {base} with {stem} ({recur_desc}) matched {dbs}")
    if len(dbs) > 1:
        raise RuntimeError(errmsg.format(dbs))
    elif len(dbs) == 1:
        # found the match!
        return dbs[0]
    else:
        # if we werent trying to search recursively, try a recursive search as a fallback
        if not recursive:
            return _handle_glob(base, stem, recursive=True)
        else:
            raise RuntimeError(f"Could not find database, using '{base}' and '{stem}'")


def _handle_path(
    pathmap: Dict[str, PathIsh],
    browser_name: str,
    *,
    key: Optional[str] = None,
    default_behaviour: str = "linux",
) -> Path:
    """
    Handles the repetitive task of having to resolve/expand a path
    which describes the location of the data directory on each
    opreating system
    """
    loc: PathIsh
    # if the user didn't provide a key, assume this is a 'sys.platform' map - using
    # darwin/linux to specify the location
    if key is None:
        key = sys.platform
    # use the key provided, or the first item (dicts after python3.7 are ordered)
    # in the pathmap if that doesnt exist
    maybeloc: Optional[PathIsh] = pathmap.get(key)
    if maybeloc is not None:
        loc = maybeloc
    else:
        warnings.warn(
            f"""Not sure where {browser_name} history is installed on {sys.platform}
Defaulting to {default_behaviour} behaviour...

If you're using a browser/platform this currently doesn't support, please make an issue
at https://github.com/seanbreckenridge/browserexport/issues/new with information.
In the meantime, you can point this directly at a history database using the --path flag"""
        )
        loc = pathmap[list(pathmap.keys())[0]]
    return expand_path(loc)


def test_handle_path() -> None:

    import pytest
    import sys

    oldplatform = sys.platform

    sys.platform = "linux"

    from .firefox import Firefox

    expected_linux = Path("~/.mozilla/firefox/").expanduser().absolute()

    assert Firefox.data_directory() == expected_linux

    sys.platform = "darwin"
    assert str(Firefox.data_directory()) == str(
        Path("~/Library/Application Support/Firefox/Profiles/").expanduser().absolute()
    )

    sys.platform = "something else"
    # should default to linux
    with pytest.warns(UserWarning, match=r"history is installed"):
        assert Firefox.data_directory() == expected_linux

    sys.platform = oldplatform
